/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.editingsupport;

import static org.eclipse.emf.ecore.ETypedElement.UNBOUNDED_MULTIPLICITY;
import static org.eclipse.emf.ecore.ETypedElement.UNSPECIFIED_MULTIPLICITY;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.DialogCellEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import org.mihalis.opal.itemSelector.DLItem;
import org.mihalis.opal.itemSelector.DualList;
import org.mihalis.opal.itemSelector.SelectionChangeEvent;
import org.mihalis.opal.itemSelector.SelectionChangeListener;

/**
 * {@link CellEditor} that opens a dialog for the selection of multiple elements.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D45A2BB67FDB88B44CADF1FE266167A5
 */

public class MultiSelectionCellEditor extends DialogCellEditor {

	/** Lower multiplicity bound. */
	int lowerMultiplicityBound;

	/** Upper multiplicity bound. */
	int upperMultiplicityBound;

	/** Mapping: label shown in GUI -> actual value. */
	private LabelValueMapping labelValueMapping;

	/** The collection of elements selected using this {@link MultiSelectionCellEditor}. */
	private Collection<?> selectedElements = CollectionUtils.emptyList();

	/** Constructors a new {@link MultiSelectionCellEditor}. */
	public MultiSelectionCellEditor(Composite parent) {
		super(parent);
	}

	/**
	 * Updates the data shown in this {@link MultiSelectionCellEditor}.
	 * 
	 * @param labelValueMapping
	 *            Mapping: label shown in GUI -> actual value.
	 * @param selectedElements
	 *            Current collection of selected elements
	 * @param lowerMultiplicityBound
	 *            Lower multiplicity bound
	 * @param upperMultiplicityBound
	 *            Upper multiplicity bound
	 */
	public void updateData(LabelValueMapping labelValueMapping, Collection<?> selectedElements,
			int lowerMultiplicityBound, int upperMultiplicityBound) {
		this.labelValueMapping = labelValueMapping;
		this.selectedElements = selectedElements;
		this.lowerMultiplicityBound = lowerMultiplicityBound;
		this.upperMultiplicityBound = upperMultiplicityBound;
	}

	/** {@inheritDoc} */
	@Override
	protected Object openDialogBox(Control cellEditorWindow) {

		MultiSelectionDialog selectionDialog = new MultiSelectionDialog(null);
		selectionDialog.open();

		return selectedElements;
	}

	/** Extension of {@link DualList} that is initialized from {@link LabelValueMapping}s. */
	private class LabelValueMappingDualList extends DualList {

		/**
		 * Constructs a new {@link LabelValueMappingDualList}.
		 * 
		 * @param parent
		 *            Parent {@link Composite} widget
		 * @param selectedElements
		 *            Collection of elements that are selected initially
		 * @param labelValueMapping
		 *            {@link LabelValueMapping} required to populate this
		 *            {@link LabelValueMappingDualList}.
		 */
		public LabelValueMappingDualList(Composite parent, Collection<?> selectedElements,
				LabelValueMapping labelValueMapping) {

			super(parent, SWT.NONE);

			List<String> labels = new ArrayList<String>();
			labels.addAll(labelValueMapping.getLabels());

			for(Object selectedItem : selectedElements) {
				String label = labelValueMapping.getLabelForValue(selectedItem);
				DLItem item = new DLItem(label);
				add(item);
				selectDoNotFireEvent(0);
				remove(item);
				labels.remove(label);
			}

			for(String label : labels) {
				DLItem item = new DLItem(label);
				add(item);
			}
		}

		/** {@inheritDoc} */
		@Override
		public void deselectAll(final boolean shouldFireEvents) {
			super.deselectAll(shouldFireEvents);

			// Required to issue a selection changed event after all elements have actually been
			// deselected. This is required for the invocation of
			// MultiSelectionDialog.updateOkButton() to work correctly.
			super.deselectAll(false);
		}
	}

	/** Shows a dialog to that allows to select multiple values. */
	private class MultiSelectionDialog extends Dialog {

		/** Element selection dialog */
		private DualList dl = null;

		/** Constructs a new {@link MultiSelectionDialog}. */
		public MultiSelectionDialog(Shell parentShell) {
			super(parentShell);

			setShellStyle(getShellStyle() | SWT.RESIZE);
		}

		/** {@inheritDoc} */
		@Override
		protected Composite createDialogArea(Composite parent) {
			/* Generate the dialog and its layout */
			Composite area = (Composite)super.createDialogArea(parent);
			area.setLayout(new FillLayout());

			dl = new LabelValueMappingDualList(area, selectedElements, labelValueMapping);

			return area;
		}

		/** {@inheritDoc} */
		@Override
		protected Control createButtonBar(Composite parent) {
			Control ctrl = super.createButtonBar(parent);

			updateOkButton();

			dl.addSelectionChangeListener(new SelectionChangeListener() {
				@Override
				public void widgetSelected(SelectionChangeEvent e) {
					updateOkButton();
				}
			});

			return ctrl;
		}

		/** Enables the Ok button iff the multiplicity constraints are met. */
		private void updateOkButton() {
			Button okButton = getButton(IDialogConstants.OK_ID);

			int n = dl.getSelectionCount();
			boolean lowerBoundOk =
					lowerMultiplicityBound == UNSPECIFIED_MULTIPLICITY ||
							n >= lowerMultiplicityBound;
			boolean upperBoundOk =
					upperMultiplicityBound == UNSPECIFIED_MULTIPLICITY ||
							upperMultiplicityBound == UNBOUNDED_MULTIPLICITY ||
							n <= upperMultiplicityBound;
			okButton.setEnabled(lowerBoundOk && upperBoundOk);
		}

		/** Helper to select the right number for "element". */
		private String titleElementStringHelper(int n) {
			return n != 1 ? "elements" : "element";
		}

		/** {@inheritDoc} */
		@Override
		protected void configureShell(Shell newShell) {
			super.configureShell(newShell);
			String title;
			if(lowerMultiplicityBound == UNSPECIFIED_MULTIPLICITY || lowerMultiplicityBound == 0) {
				if(upperMultiplicityBound == UNSPECIFIED_MULTIPLICITY ||
						upperMultiplicityBound == UNBOUNDED_MULTIPLICITY) {
					title = "Select multiple elements";
				} else {
					title =
							"Select at most " + upperMultiplicityBound + " " +
									titleElementStringHelper(upperMultiplicityBound);
				}
			} else {
				if(upperMultiplicityBound == UNSPECIFIED_MULTIPLICITY ||
						upperMultiplicityBound == UNBOUNDED_MULTIPLICITY) {
					title =
							"Select at least " + lowerMultiplicityBound + " " +
									titleElementStringHelper(lowerMultiplicityBound);
				} else {
					title =
							"Select " + lowerMultiplicityBound + " to " + upperMultiplicityBound +
									" " + titleElementStringHelper(upperMultiplicityBound);
				}
			}
			newShell.setText(title);
		}

		/** {@inheritDoc} */
		@Override
		protected Point getInitialSize() {
			return new Point(800, 400);
		}

		/** Returns the set of selected elements. */
		private Collection<Object> getSelectedItems() {
			Collection<Object> rval = new ArrayList<Object>();
			for(DLItem item : dl.getSelection()) {
				if(item != null) {
					rval.add(labelValueMapping.getValueForLabel(item.getText()));
				}
			}
			return rval;
		}

		/** {@inheritDoc} */
		@Override
		protected void okPressed() {
			// In case the user presses ok, update the corresponding field of
			// MultiSelectionCellEditor that wraps this MultiSelectionDialog
			selectedElements = getSelectedItems();

			super.okPressed();
		}
	}
}
