/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.dialog;

import static org.eclipse.jface.dialogs.IDialogConstants.OK_ID;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;

/**
 * This dialog offers a common dialog to show multi-root tree structures and
 * allows user to select one item in the tree. It also offers interface to edit
 * the tree on the fly.
 * 
 * @author mou
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 8F165DEFB14FA23E90802923896C1142
 */
public class ElementTreeSingleSelectDialog extends TitleAreaDialog {

	/** the editing module */
	private IDialogEditSupport dialogEditSupport;

	/** the root elements */
	private ArrayList<Object> rootElements;

	/** the selected element as tree path */
	private TreePath selectedPath;

	/** the content provider */
	private ITreeContentProvider contentSupport;

	/** the label provider */
	private ILabelProvider labelSupport;

	/** The gui composite */
	private ElementTreeSingleSelectDialogGUI gui;

	/** The title of the dialog. */
	private String dialogTitle = "Select";

	/** The message that explains what should be selected. */
	private String selectionMessage = "Select an element";

	/**
	 * Main constructor
	 * 
	 * @param parentShell
	 *            the parent shell
	 * @param roots
	 *            the root elements
	 * @param selected
	 *            the pre-selected element or null
	 * @param contentProvider
	 *            the content provider
	 * @param labelProvider
	 *            the label provider
	 * @param edit
	 *            the editing module or null
	 */
	public ElementTreeSingleSelectDialog(Shell parentShell, List<?> roots, Object selected,
			ITreeContentProvider contentProvider, ILabelProvider labelProvider,
			IDialogEditSupport edit) {
		super(parentShell);

		if(roots == null) {
			rootElements = new ArrayList<Object>();
		} else {
			rootElements = new ArrayList<Object>(roots);
		}

		dialogEditSupport = edit;
		contentSupport = contentProvider;
		labelSupport = labelProvider;

		if(selected == null) {
			selectedPath = null;
		} else {
			selectedPath = computeTreePath(rootElements, selected);
		}
	}

	/**
	 * Constructor.
	 * 
	 * @param parentShell
	 *            the parent shell
	 * @param dialogTitle
	 *            the title of the dialog
	 * @param selectionMessage
	 *            the message explaining what is the selection about
	 * @param roots
	 *            the root elements
	 * @param selected
	 *            the pre-selected element or null
	 * @param contentProvider
	 *            the content provider
	 * @param labelProvider
	 *            the label provider
	 * @param edit
	 *            the editing module or null
	 */
	public ElementTreeSingleSelectDialog(Shell parentShell, String dialogTitle,
			String selectionMessage, List<?> roots, Object selected,
			ITreeContentProvider contentProvider, ILabelProvider labelProvider,
			IDialogEditSupport edit) {
		this(parentShell, roots, selected, contentProvider, labelProvider, edit);
		this.dialogTitle = dialogTitle;
		this.selectionMessage = selectionMessage;
	}

	/** {@inheritDoc} */
	@Override
	protected Control createDialogArea(Composite parent) {
		setHelpAvailable(false);

		setTitle(dialogTitle);
		setMessage(selectionMessage);

		// create controls
		Composite area = (Composite)super.createDialogArea(parent);
		gui = new ElementTreeSingleSelectDialogGUI(area, SWT.NONE);
		gui.setLayoutData(new GridData(GridData.FILL_BOTH));

		// hide edit toolbar if no edit function is available
		if(dialogEditSupport == null) {
			gui.getLeftToolBar().setVisible(false);
			((GridData)gui.getLeftToolBar().getLayoutData()).exclude = true;
		}

		// label, content and input
		gui.getElementTreeViewer().setLabelProvider(labelSupport);

		gui.getElementTreeViewer().setContentProvider(
				new MultiRootTreeContentProvider(rootElements, contentSupport));

		gui.getElementTreeViewer().setInput(rootElements);

		if(selectedPath != null)
			gui.getElementTreeViewer().setSelection(new TreeSelection(selectedPath), true);

		// create listeners
		gui.getCreateButton().addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				createElement();
			}
		});

		gui.getDeleteButton().addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				deleteElement();
			}
		});

		gui.getEditButton().addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				editElement();
			}
		});

		gui.getElementTreeViewer().addPostSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				selectElement();
			}
		});

		return area;
	}

	/** {@inheritDoc} */
	@Override
	protected Point getInitialSize() {
		return new Point(480, 480);
	}

	/** Returns configuration. */
	public IDialogEditSupport getEditSupport() {
		return dialogEditSupport;
	}

	/** Returns rootElements. */
	public List<Object> getRootElements() {
		return rootElements;
	}

	/** Returns selectedElements. */
	public Object getSelectedElement() {
		if(selectedPath == null)
			return null;
		return selectedPath.getLastSegment();
	}

	/** Returns gui. */
	public ElementTreeSingleSelectDialogGUI getGui() {
		return gui;
	}

	/** Returns contentSupport. */
	public IStructuredContentProvider getContentProvider() {
		return contentSupport;
	}

	/** Returns labelSupport. */
	public ILabelProvider getLabelProvider() {
		return labelSupport;
	}

	/** Create an element */
	private void createElement() {
		if(dialogEditSupport == null)
			return;

		try {

			// find the path to hold the new element
			ITreeSelection selection = (ITreeSelection)gui.getElementTreeViewer().getSelection();
			if(selection.getPaths() != null && selection.getPaths().length != 1)
				return; // at most 1 element can be selected

			// create the new element
			Object o;
			if(selection.getPaths() == null) {
				o = dialogEditSupport.createElement(null);
			} else {
				o = dialogEditSupport.createElement(selection.getPaths()[0].getLastSegment());
			}
			if(o == null)
				return;

			TreePath path;
			if(selection.getPaths() == null) {
				path = new TreePath(new Object[] {rootElements});
				rootElements.add(o);
			} else {
				path = selection.getPaths()[0];
			}

			// insert the new element
			gui.getElementTreeViewer().add(path, o);
			gui.getElementTreeViewer().expandToLevel(path, 1);
			gui.getElementTreeViewer().reveal(o);
			gui.getElementTreeViewer().refresh();

			TreeSelection sel = new TreeSelection(path.createChildPath(o));
			gui.getElementTreeViewer().setSelection(sel);
		} catch(Exception e2) {
			e2.printStackTrace();
		}
	}

	/** Call delete function on the selected elements */
	private void deleteElement() {
		if(dialogEditSupport == null)
			return;

		try {
			ITreeSelection selection = (ITreeSelection)gui.getElementTreeViewer().getSelection();
			if(selection.getPaths() == null)
				return;

			// delete selected elements
			for(TreePath path : selection.getPaths()) {
				if(dialogEditSupport.deleteElement(path.getLastSegment())) {
					if(rootElements.contains(path.getLastSegment()))
						rootElements.remove(path.getLastSegment());

					gui.getElementTreeViewer().remove(path);

					if(path.equals(selectedPath)) {
						selectedPath = null;
					}
				}
			}
		} catch(Exception e2) {
			e2.printStackTrace();
		}
	}

	/** Call edit function on the selected element */
	private void editElement() {
		if(dialogEditSupport == null)
			return;

		try {
			// find the path to the element
			ITreeSelection selection = (ITreeSelection)gui.getElementTreeViewer().getSelection();
			if(selection.getPaths() == null || selection.getPaths().length != 1)
				return; // 1 element must be selected

			TreePath path = selection.getPaths()[0];
			dialogEditSupport.editElement(path.getLastSegment());

			gui.getElementTreeViewer().refresh();
		} catch(Exception e2) {
			e2.printStackTrace();
		}
	}

	/** update the selected element by its tree path */
	private void selectElement() {
		// find the path to the element
		ITreeSelection selection = (ITreeSelection)gui.getElementTreeViewer().getSelection();
		if(selection.getPaths() == null || selection.getPaths().length != 1 ||
				!acceptElement(selection.getPaths()[0].getLastSegment())) {
			selectedPath = null;
			getButton(OK_ID).setEnabled(false);
		} else {
			selectedPath = selection.getPaths()[0];
			getButton(OK_ID).setEnabled(true);
		}
	}

	/**
	 * This method can be overwritten by subclasses to accept only elements of a
	 * certain kind. For not-accepted elements the OK-field is disabled.
	 * 
	 * The default implementation accepts all elements.
	 * 
	 * @param element
	 *            the element which is currently selected
	 * 
	 * @return Whether the {@code element} is accepted.
	 */
	public boolean acceptElement(Object element) {
		return true;
	}

	/**
	 * Create the tree path from one of the given root elements to the child
	 * element. The content provider will be used to compute the direct parent
	 * of the given child element.
	 * 
	 * @param roots
	 *            all possible parent elements
	 * @param leaf
	 *            the child element
	 * @return a tree path starting from one of the root elements and up to the
	 *         child element, or null if no such path exists
	 */
	private TreePath computeTreePath(List<?> roots, Object leaf) {
		LinkedList<Object> path = new LinkedList<Object>();

		Object parent = leaf;
		while(!roots.contains(parent)) {
			path.addFirst(parent);
			parent = contentSupport.getParent(parent);

			if(parent == null)
				return null;
		}
		path.addFirst(parent);

		return new TreePath(path.toArray());
	}
}
