/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.eclipse.core.runtime.IStatus.ERROR;
import static org.eclipse.jface.dialogs.ErrorDialog.openError;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.activateConfiguration;
import static org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils.deactivateConfiguration;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.DEFAULT_CONFIGURATION_NAME;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConfigurationsTransitively;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConstraintsTransitively;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.retrieveDefaultConfig;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.CheckStateChangedEvent;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.ICheckStateListener;
import org.eclipse.jface.viewers.ICheckStateProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.TreeColumn;
import org.fortiss.tooling.base.model.element.ConstraintConfiguration;
import org.fortiss.tooling.base.model.element.ElementFactory;
import org.fortiss.tooling.base.model.element.IConstraintBasedProcess;
import org.fortiss.tooling.base.ui.ToolingBaseUIActivator;
import org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.service.IConstraintService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.extension.base.EditorBase;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IActionService;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Editor for {@link IConstraintBasedProcess}s.
 * 
 * TODO(2931): should not be an editor but a widget
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 1B0468C1E3B5A7BDDA69C2E5D2F23116
 */
public class ConstraintBasedProcessEditor<CBP extends IConstraintBasedProcess> extends
		EditorBase<CBP> implements ContextMenuContextProvider, ISelectionProvider {

	/** Model context of the edited project. */
	private ITopLevelElement top;

	/** The default configuration. */
	private ConstraintConfiguration defaultConfig;

	/** Constraint container for the edited project. */
	private IConstraintInstanceContainer cstrContainer;

	/** Tree viewer for the configurations. */
	private CheckboxTreeViewer treeViewer;

	/** Selection of the tree. */
	private ISelection selection;

	/** Comboviewer for the current objective. */
	private ComboViewer currentObjective;

	/** Adapter for refreshing content. */
	private EContentAdapter refreshAdapter;

	/** Reference to a constraint configuration. */
	private static class WithContextualConfiguration {

		/** Contextual constraint configuration. */
		ConstraintConfiguration config;

		/** Constructor. */
		public WithContextualConfiguration(ConstraintConfiguration config) {
			this.config = config;
		}
	}

	/** Constraint accompanied with a configuration providing the context. */
	private static class ConstraintWithContextualConfiguration extends WithContextualConfiguration {

		/** Constraint. */
		Class<? extends IConstraint> cstr;

		/** Constructor. */
		public ConstraintWithContextualConfiguration(Class<? extends IConstraint> cstr,
				ConstraintConfiguration config) {
			super(config);
			this.cstr = cstr;
		}
	}

	/**
	 * Configuration accompanied with another configuration providing the context.
	 * Intent is that the contained configuration is a dependency of the context one.
	 */
	private static class ConfigurationRefWithContextualConfiguration extends
			WithContextualConfiguration {

		/** Constraint. */
		ConstraintConfiguration target;

		/** Constructor. */
		public ConfigurationRefWithContextualConfiguration(ConstraintConfiguration configRef,
				ConstraintConfiguration config) {
			super(config);
			this.target = configRef;
		}
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		top = IPersistencyService.getInstance().getTopLevelElementFor(editedObject);
		getSite().setSelectionProvider(this);
		defaultConfig = retrieveDefaultConfig(editedObject);
		cstrContainer = editedObject.getConstraintInstanceContainer();
		parent.setLayout(new GridLayout(1, true));
		createTree(parent);
		MenuManager menuManager = IContextMenuService.getInstance().createDefaultContextMenu(this);
		Menu menu = menuManager.createContextMenu(treeViewer.getTree());
		treeViewer.getTree().setMenu(menu);
		createCurrentObjectivePart(parent);

		refreshAdapter = new EContentAdapter() {

			@Override
			public void notifyChanged(Notification notification) {
				treeViewer.refresh();
				currentObjective.refresh();
			}
		};
		editedObject.eAdapters().add(refreshAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		EcoreUtils.disableNotificationsOf(editedObject);
		editedObject.eAdapters().remove(refreshAdapter);
		EcoreUtils.enableNotificationsOf(editedObject);
		super.dispose();
	}

	/** Create current the controls "Current objective:". */
	private void createCurrentObjectivePart(Composite parent) {
		Composite currentObjComposite = new Composite(parent, SWT.None);
		currentObjComposite.setLayout(new GridLayout(2, false));
		Label currentObjLabel = new Label(currentObjComposite, SWT.None);
		currentObjLabel.setText("Current objective:");
		currentObjective = new ComboViewer(currentObjComposite, SWT.READ_ONLY);
		currentObjective.setContentProvider(ArrayContentProvider.getInstance());
		currentObjective.setLabelProvider(new LabelProvider() {
			@Override
			public String getText(Object elt) {
				return ((ConstraintConfiguration)elt).getName();
			}
		});
		currentObjective.setInput(editedObject.getConfigurations());
		if(editedObject.getCurrentObjective() == null) {
			top.runAsNonDirtyingCommand(() -> editedObject.setCurrentObjective(defaultConfig));
		}
		currentObjective.setSelection(new StructuredSelection(editedObject.getCurrentObjective()));
		currentObjective.addPostSelectionChangedListener(new ISelectionChangedListener() {

			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				if(!(event.getSelection() instanceof StructuredSelection)) {
					return;
				}
				final ConstraintConfiguration oldObjective = editedObject.getCurrentObjective();
				ConstraintConfiguration newObjective =
						(ConstraintConfiguration)((StructuredSelection)event.getSelection())
								.getFirstElement();
				top.runAsCommand(() -> {
					ConstraintConfiguration effectiveNewObjective =
							newObjective == null ? defaultConfig : newObjective;
					editedObject.setCurrentObjective(effectiveNewObjective);
					List<ConstraintConfiguration> activeConfigs =
							getActiveConfigurationsTransitively(effectiveNewObjective, null);
					if(!activeConfigs.contains(oldObjective)) {
						deactivateConfiguration(oldObjective, cstrContainer, effectiveNewObjective);
					}
					activateConfiguration(effectiveNewObjective, cstrContainer);
				});
			}
		});
	}

	/** Fake child used to add new elements. */
	private static class FakeChild {
		// Nothing to implement: just a marker class
	}

	/** Content provider for the configuration tree. */
	private class ConfigTreeContentProvider extends TreeContentProviderBase {
		/** {@inheritDoc} */
		@Override
		public Object[] getChildren(Object parentElement) {
			if(parentElement instanceof IConstraintBasedProcess) {
				IConstraintBasedProcess cbdp = (IConstraintBasedProcess)parentElement;
				ArrayList<Object> configs = new ArrayList<>();
				configs.add(new FakeChild());
				configs.addAll(cbdp.getConfigurations());
				return configs.toArray();
			} else if(parentElement instanceof ConstraintConfiguration) {
				List<Object> types = new ArrayList<Object>();
				Set<Class<? extends IConstraint>> availableCstrs =
						IConstraintUIService.getInstance().getAllConstraints();
				availableCstrs.removeAll(IConstraintUIService.getInstance()
						.getAlwaysActivatedConstraints());
				types.addAll(availableCstrs
						.stream()
						.map(c -> new ConstraintWithContextualConfiguration(c,
								(ConstraintConfiguration)parentElement))
						.collect(Collectors.toList()));
				if(editedObject.getConfigurations().size() > 1) {
					List<ConfigurationRefWithContextualConfiguration> refs =
							editedObject
									.getConfigurations()
									.stream()
									.map(t -> new ConfigurationRefWithContextualConfiguration(t,
											(ConstraintConfiguration)parentElement))
									.collect(Collectors.toList());
					types.add(refs);
				}
				return types.toArray();
			} else if(parentElement instanceof List<?>) {
				return ((List<?>)parentElement).toArray();
			}
			return new Object[0];
		}
	}

	/** Text labels for the tree of configuratinos. */
	private String getLabelText(Object element) {
		if(element instanceof FakeChild) {
			return "click to add a new configuration";
		} else if(element instanceof ConstraintConfiguration) {
			return ((ConstraintConfiguration)element).getName();
		} else if(element instanceof ConstraintWithContextualConfiguration) {
			IConstraintUIService cuis = IConstraintUIService.getInstance();
			return cuis.getDescription(((ConstraintWithContextualConfiguration)element).cstr);
		} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
			return ((ConfigurationRefWithContextualConfiguration)element).target.getName();
		} else if(element instanceof List<?>) {
			return "Required configurations";
		}
		return null;
	}

	/** Check state provider for the tree of configurations. */
	private class ConfigTreeCheckStateProvider implements ICheckStateProvider {

		/** {@inheritDoc} */
		@Override
		public boolean isGrayed(Object element) {
			return !(element instanceof ConstraintWithContextualConfiguration || element instanceof ConfigurationRefWithContextualConfiguration);
		}

		/** {@inheritDoc} */
		@Override
		public boolean isChecked(Object element) {
			if(element instanceof ConstraintWithContextualConfiguration) {
				ConstraintWithContextualConfiguration cstrInContext =
						(ConstraintWithContextualConfiguration)element;
				IConstraintService cs = IConstraintService.getInstance();
				String name = cs.getName(cstrInContext.cstr);
				return cstrInContext.config.getActiveConstraints().contains(name);
			} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
				ConfigurationRefWithContextualConfiguration configRef =
						(ConfigurationRefWithContextualConfiguration)element;
				return configRef.config.getIncludedConfigurations().contains(configRef.target);
			}
			return true;
		}
	}

	/** Check state listener for the tree of configurations. */
	private class ConfigTreeCheckStateListener implements ICheckStateListener {
		/** {@inheritDoc} */
		@Override
		public void checkStateChanged(CheckStateChangedEvent event) {
			Object element = event.getElement();
			if(element instanceof ConstraintWithContextualConfiguration) {
				ConstraintWithContextualConfiguration cstrInContext =
						(ConstraintWithContextualConfiguration)element;
				IConstraintService cs = IConstraintService.getInstance();
				String name = cs.getName(cstrInContext.cstr);
				top.runAsCommand(() -> {
					IConstraintUIService cuis = IConstraintUIService.getInstance();
					ConstraintConfiguration currentObj = editedObject.getCurrentObjective();
					if(event.getChecked()) {
						cstrInContext.config.getActiveConstraints().add(name);
						List<ConstraintConfiguration> activeConfigs =
								getActiveConfigurationsTransitively(currentObj, null);
						if(activeConfigs.contains(cstrInContext.config)) {
							cuis.activate(cstrInContext.cstr, cstrContainer);
						}
					} else {
						cstrInContext.config.getActiveConstraints().remove(name);
						List<Class<? extends IConstraint>> stillActiveCstrs =
								getActiveConstraintsTransitively(currentObj, cstrInContext.config);
						if(!stillActiveCstrs.contains(cstrInContext.cstr)) {
							cuis.deactivate(cstrInContext.cstr, cstrContainer);
						}
					}
				});
			} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
				ConfigurationRefWithContextualConfiguration configRef =
						(ConfigurationRefWithContextualConfiguration)element;
				top.runAsCommand(() -> {
					if(event.getChecked()) {
						configRef.config.getIncludedConfigurations().add(configRef.target);
						ConstraintsBaseUIUtils.activateConfiguration(configRef.target,
								cstrContainer);
					} else {
						configRef.config.getIncludedConfigurations().remove(configRef.target);
						ConstraintsBaseUIUtils.deactivateConfiguration(configRef.target,
								cstrContainer, configRef.config);
					}
				});
			}
		}
	}

	/** Tree for configurations. */
	private class ConfigTree extends CheckboxTreeViewer implements ISelectionChangedListener {

		/** Constructor. */
		public ConfigTree(Composite parent) {
			super(parent);
			getTree().setLayoutData(new GridData(800, 300));
			setContentProvider(new ConfigTreeContentProvider());
			setCheckStateProvider(new ConfigTreeCheckStateProvider());
			addCheckStateListener(new ConfigTreeCheckStateListener());
			addSelectionChangedListener(this);
		}

		/** {@inheritDoc} */
		@Override
		public void selectionChanged(SelectionChangedEvent event) {
			selection = event.getSelection();
			IActionService.getInstance().refresh();
		}
	}

	/** Creates the tree of configurations. */
	private void createTree(Composite parent) {
		treeViewer = new ConfigTree(parent);
		TreeColumn column = new TreeColumn(treeViewer.getTree(), SWT.LEFT);
		column.setText("Constraint configurations");
		column.setWidth(800);

		TreeViewerColumn onlyViewerColumn = new TreeViewerColumn(treeViewer, column);
		onlyViewerColumn.setLabelProvider(new ColumnLabelProvider() {

			/** {@inheritDoc} */
			@Override
			public String getText(Object element) {
				return getLabelText(element);
			}
		});
		onlyViewerColumn.setEditingSupport(new EditingSupport(treeViewer) {

			@Override
			protected void setValue(Object element, Object value) {
				if(element instanceof FakeChild) {
					if(!alreadyExists((String)value)) {
						ConstraintConfiguration newconfig =
								ElementFactory.eINSTANCE.createConstraintConfiguration();
						newconfig.setName((String)value);
						top.runAsCommand(() -> IElementCompositorService.getInstance().compose(
								editedObject, newconfig, null));
					}
				}
				if(element instanceof ConstraintConfiguration) {
					ConstraintConfiguration config = (ConstraintConfiguration)element;
					if(config.getName().equals(DEFAULT_CONFIGURATION_NAME) ||
							config.getName().equals(value)) {
						return;
					}
					if(!alreadyExists((String)value)) {
						top.runAsCommand(() -> ((ConstraintConfiguration)element)
								.setName((String)value));
					}
				}
			}

			@Override
			protected Object getValue(Object element) {
				return getLabelText(element);
			}

			@Override
			protected CellEditor getCellEditor(Object element) {
				return new TextCellEditor((Composite)getViewer().getControl());
			}

			@Override
			protected boolean canEdit(Object element) {
				return element instanceof ConstraintConfiguration || element instanceof FakeChild;
			}
		});
		treeViewer.setComparator(new ConstraintComparator());
		treeViewer.setInput(editedObject);
	}

	/**
	 * Checks if the name already exists among current configurations, if so displays an error
	 * message and return true, otherwise returns false.
	 */
	private boolean alreadyExists(String name) {
		List<String> configNames =
				editedObject.getConfigurations().stream().map(c -> c.getName())
						.collect(Collectors.toList());
		if(configNames.contains(name)) {
			Status status =
					new Status(ERROR, ToolingBaseUIActivator.PLUGIN_ID, 0,
							"A configuration with this name already exists.", null);
			openError(Display.getCurrent().getActiveShell(), "Error", null, status);
			return true;
		}
		return false;
	}

	/** Comparator for the contents of the tree. */
	private class ConstraintComparator extends ViewerComparator {

		/** {@inheritDoc} */
		@Override
		public int compare(Viewer viewer, Object e1, Object e2) {
			if(e1 instanceof FakeChild) {
				return -1;
			} else if(e2 instanceof FakeChild) {
				return 1;
			} else if(e1 instanceof ConstraintConfiguration &&
					e2 instanceof ConstraintConfiguration) {
				return ((ConstraintConfiguration)e1).getName().compareToIgnoreCase(
						((ConstraintConfiguration)e2).getName());
			} else if(e1 instanceof ConstraintWithContextualConfiguration &&
					e2 instanceof ConstraintWithContextualConfiguration) {
				IConstraintUIService cuis = IConstraintUIService.getInstance();
				String desc1 =
						cuis.getDescription(((ConstraintWithContextualConfiguration)e1).cstr);
				String desc2 =
						cuis.getDescription(((ConstraintWithContextualConfiguration)e2).cstr);
				return desc1.compareToIgnoreCase(desc2);
			} else if(e1 instanceof ConfigurationRefWithContextualConfiguration &&
					e2 instanceof ConfigurationRefWithContextualConfiguration) {
				String n1 = ((ConfigurationRefWithContextualConfiguration)e1).target.getName();
				String n2 = ((ConfigurationRefWithContextualConfiguration)e2).target.getName();
				return n1.compareToIgnoreCase(n2);
			}
			return super.compare(viewer, e1, e2);
		}
	}

	/** {@inheritDoc} */
	@Override
	public EObject getSelectedModelElement() {
		if(getSelectedModelElementList().isEmpty()) {
			return null;
		}
		// Returns the first element, which is ensured to exist by the previous statement.
		Object selected = getSelectedModelElementList().get(0);
		return selected instanceof EObject ? (EObject)selected : null;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public List<EObject> getSelectedModelElementList() {
		boolean isStructured = selection instanceof StructuredSelection;
		return isStructured ? ((StructuredSelection)selection).toList() : new ArrayList<>();
	}

	/** {@inheritDoc} */
	@Override
	public void addSelectionChangedListener(ISelectionChangedListener listener) {
		// Not supported
	}

	/** {@inheritDoc} */
	@Override
	public ISelection getSelection() {
		return selection == null ? new StructuredSelection() : selection;
	}

	/** {@inheritDoc} */
	@Override
	public void removeSelectionChangedListener(ISelectionChangedListener listener) {
		// Not supported
	}

	/** {@inheritDoc} */
	@Override
	public void setSelection(ISelection selection) {
		// Not supported
	}
}
