/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.eclipse.jface.viewers.AbstractTreeViewer.ALL_LEVELS;
import static org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils.getEObjectElements;

import java.util.Collection;
import java.util.List;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.fortiss.tooling.base.ui.dnd.jface.ViewerElementCompositionDropAdapter;
import org.fortiss.tooling.kernel.ui.extension.base.EditorBase;
import org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils;

/**
 * An editor which only consists of a tree viewer. It uses {@link TreeViewer},
 * which also allows to define columns, i.e., use the tree and table at the same
 * time.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 88EBBD15A1CEB3F0501DEE1756334D4D
 */
public abstract class TreeViewerEditorBase<T extends EObject> extends EditorBase<T> {

	/** Stores the {@link TreeViewer}. */
	private TreeViewer treeViewer;

	/**
	 * Flag if redraw of the underlying {@link #treeViewer} is enabled. Redrawing is initially
	 * enabled in {@link #createPartControl(Composite)}, but it can be temporarily disabled in order
	 * to speed up long running operations that modify the data displayed in this
	 * {@link TreeViewerEditorBase}. Otherwise, constant updates of the tree view significantly slow
	 * down the computation (often the GUI updates consume more time than the actual operation).
	 */
	boolean redraw;

	/** The adapter used for dealing with changes of the content. */
	private final Adapter editedObjectChanged = new EContentAdapter() {
		@Override
		public void notifyChanged(Notification notification) {
			super.notifyChanged(notification);
			if(notification.getEventType() != Notification.REMOVING_ADAPTER) {
				Display.getDefault().asyncExec(() -> {
					if(!treeViewer.getTree().isDisposed() && getRedraw()) {
						treeViewer.refresh(true);
					}
				});
			}
		}
	};

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		treeViewer =
				new TreeViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER |
						SWT.FULL_SELECTION);
		redraw = true;

		setupTreeViewer(treeViewer);

		ViewerElementCompositionDropAdapter dndAdapter =
				new ViewerElementCompositionDropAdapter(treeViewer);
		treeViewer.addDropSupport(dndAdapter.getSupportedDNDOperations(),
				dndAdapter.getPreferredTransfers(), dndAdapter);

		treeViewer.setInput(editedObject);
		getEditorSite().setSelectionProvider(treeViewer);

		editedObject.eAdapters().add(editedObjectChanged);

		getEditorSite().setSelectionProvider(treeViewer);
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		editedObject.eAdapters().remove(editedObjectChanged);
		super.dispose();
	}

	/** Returns the elements currently selected in the tree. */
	protected Collection<EObject> getTreeSelection() {
		List<EObject> result = getEObjectElements(treeViewer.getSelection());
		if(result.isEmpty()) {
			result.add(editedObject);
		}
		return result;
	}

	/**
	 * This method is used to configure the tree viewer, e.g. set the content
	 * provider, and the label provider.
	 */
	protected abstract void setupTreeViewer(TreeViewer treeViewer);

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		treeViewer.getTree().setFocus();
	}

	/** Returns {@link #treeViewer}. */
	public TreeViewer getTreeViewer() {
		return treeViewer;
	}

	/** Sets {@link #redraw}. */
	public void setRedraw(boolean redraw) {
		this.redraw = redraw;
		if(treeViewer.getTree().isDisposed()) {
			return;
		}
		treeViewer.getTree().setRedraw(redraw);
		if(redraw) {
			treeViewer.refresh(true);
		}
	}

	/** Returns {@link #redraw}. */
	boolean getRedraw() {
		return redraw;
	}

	/** Expands the selected sub-tree in a {@link TreeViewerEditorBase}. */
	public class ExpandSubTreeAction extends Action {

		/** Editor in which a sub-tree should be expanded by this action. */
		private TreeViewerEditorBase<?> editor;

		/** Constructor. */
		public ExpandSubTreeAction(TreeViewerEditorBase<?> editor) {
			super("Expand");
			setToolTipText("Expands the selected (sub-tree)");
			this.editor = editor;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			ISelection selection = EObjectSelectionUtils.getCurrentSelection();

			if(selection instanceof IStructuredSelection && !selection.isEmpty()) {
				Object object = ((IStructuredSelection)selection).getFirstElement();
				if(object != null) {
					editor.getTreeViewer().expandToLevel(object, ALL_LEVELS);
				}
			}
		}
	}

	/** Collapses the selected sub-tree in a {@link TreeViewerEditorBase}. */
	public class CollapseSubTreeAction extends Action {

		/** Editor in which a sub-tree should be expanded by this action. */
		private TreeViewerEditorBase<?> editor;

		/** Constructor. */
		public CollapseSubTreeAction(TreeViewerEditorBase<?> editor) {
			super("Collapse");
			setToolTipText("Collapses the selected (sub-tree)");
			this.editor = editor;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			ISelection selection = EObjectSelectionUtils.getCurrentSelection();

			if(selection instanceof IStructuredSelection && !selection.isEmpty()) {
				Object object = ((IStructuredSelection)selection).getFirstElement();
				if(object != null) {
					editor.getTreeViewer().collapseToLevel(object, ALL_LEVELS);
				}
			}
		}
	}
}
