/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.camera;

import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static java.lang.Math.toRadians;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.X_UNIT_VECTOR;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.Z_UNIT_VECTOR;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.lookAt;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.rotate;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.translate;

/**
 * A {@link Camera} which can be configured with pitch and yaw angle.
 * With both angles at 0.0 degree the camera looks along the Y axis.
 * <P>
 * This camera does not support setting the target location directly.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 1B3278814179BC357FB550B79745922F
 */
public class MovingCameraWithPitchAndYaw extends Camera {

	/** The yaw angle. */
	private float yawAngle = 0.0f;

	/** The pitch angle. */
	private float pitchAngle = 0.0f;

	/** Alter the pitch angle. */
	public void pitch(float amount) {
		pitchAngle += amount;
		if(pitchAngle >= 85.0f) {
			pitchAngle = 85.0f;
		} else if(pitchAngle <= -85.0) {
			pitchAngle = -85.0f;
		}
	}

	/** Alter the yaw angle. */
	public void yaw(float amount) {
		yawAngle += amount;
		if(yawAngle >= 360.f) {
			yawAngle = yawAngle - 360.f;
		} else if(yawAngle <= -360) {
			yawAngle = yawAngle + 360.f;
		}
	}

	/** Move forward for the given distance. */
	public void forward(float distance) {
		float dx = distance * (float)sin(toRadians(yawAngle));
		float dy = distance * (float)cos(toRadians(yawAngle));
		super.alterPosition(dx, dy, 0);
	}

	/** Move backward for the given distance. */
	public void backward(float distance) {
		forward(-distance);
	}

	/** Move left for the given distance. */
	public void left(float distance) {
		float dx = distance * (float)sin(toRadians(yawAngle - 90));
		float dy = distance * (float)cos(toRadians(yawAngle - 90));
		super.alterPosition(dx, dy, 0);
	}

	/** Move right for the given distance. */
	public void right(float distance) {
		left(-distance);
	}

	/** Move up for the given distance. */
	public void up(float distance) {
		super.alterPosition(0, 0, distance);
	}

	/** Move down for the given distance. */
	public void down(float distance) {
		up(-distance);
	}

	/** {@inheritDoc} */
	@Override
	public void applyForRendering() {
		// from origin look at Y axis
		lookAt(0, 0, 0, 0, 1.0f, 0);
		rotate(pitchAngle, X_UNIT_VECTOR);
		rotate(yawAngle, Z_UNIT_VECTOR);
		translate(-pos.x, -pos.y, -pos.z);
	}

	/** {@inheritDoc} */
	@Override
	public void applyForPicking() {
		applyForRendering();
	}

	/** {@inheritDoc} */
	@Override
	public void setTarget(float tx, float ty, float tz) {
		throw new UnsupportedOperationException();
	}

	/** {@inheritDoc} */
	@Override
	public void alterTarget(float dx, float dy, float dz) {
		throw new UnsupportedOperationException();
	}
}
