/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.draw;

import static org.fortiss.tooling.graphicsGL.ui.ToolingGraphicsGLUIActivator.getDefault;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.disableTextures;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.enableTextures;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.initializeGLfor2D;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.initializeGLfor3D;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.awt.Font;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.swt.SWT;
import org.eclipse.swt.opengl.GLCanvas;
import org.eclipse.swt.opengl.GLData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Listener;
import org.lwjgl.opengl.GL;
import org.newdawn.slick.Color;
import org.newdawn.slick.TrueTypeFont;

/**
 * Canvas for drawing LWJGL content onto a SWT widget. It uses double buffer, depth buffer and four
 * sample buffers. The canvas also provides a mechanism to draw strings, for which it loads and
 * stores the true type font instance from slick utilities.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: FFCEE98593AEFA7829436F7961C8F25A
 */
@SuppressWarnings("deprecation")
public class LwjglCanvas extends GLCanvas {

	/** Constructor. */
	public LwjglCanvas(Composite parent, GLData data) {
		super(parent, SWT.NONE, data);
	}

	/** Constructor. */
	public LwjglCanvas(Composite parent) {
		this(parent, createDefaultGLData());
	}

	/** Creates the default {@link GLData}. */
	private static GLData createDefaultGLData() {
		GLData data = new GLData();
		// use double buffering
		data.doubleBuffer = true;
		// allow depth buffering
		data.depthSize = 1;
		// enable four sample buffers for anti-aliasing
		data.sampleBuffers = 1;
		data.samples = 4;

		return data;
	}

	/** Adds the given listener to handle the canvas resize events. */
	public void addResizeHandler(Listener handler) {
		addListener(SWT.Resize, handler);
	}

	/** Sets this canvas as the current canvas and tells OGL to use it. */
	public void setCurrentAndUseContext() {
		try {
			setCurrent();
			GL.createCapabilities();
		} catch(Exception ex) {
			error(getDefault(), ex.getMessage(), ex);
		}
	}

	/** Sets the view port and perspective projection matrix. */
	public void setViewportAndPerspective(int width, int height, float fieldOfViewAngle,
			float nearDistance, float farDistance) {
		if(nearDistance <= 0 || farDistance <= 0) {
			throw new IllegalArgumentException("Z distances of perspective must be greater 0.");
		}
		float aspect = (float)width / (float)height;
		setCurrentAndUseContext();
		initializeGLfor3D(width, height, fieldOfViewAngle, aspect, nearDistance, farDistance);
	}

	/** Sets the viewport for 2D drawing (e.g. text over a 3D view). */
	public void setViewPortFor2D(int width, int height) {
		setCurrentAndUseContext();
		initializeGLfor2D(width, height);
	}

	/** Buffer for loaded fonts. */
	// using deprecated TrueTypeFont, since UnicodeFont is not yet included in slick-utils.jar
	private final Map<Font, TrueTypeFont> fontMap = new HashMap<Font, TrueTypeFont>();

	/** Ensures that fonts are ready for use. */
	private void ensureFontLoading(Font font) {
		if(font != null && !fontMap.containsKey(font)) {
			fontMap.put(font, new TrueTypeFont(font, true));
		}
	}

	/** Draws the given string at the given location with the given font. */
	public void drawString(String msg, int x, int y, Font font, Color color) {
		ensureFontLoading(font);
		enableTextures();
		fontMap.get(font).drawString(x, y, msg,
				new org.newdawn.slick.Color(color.r, color.g, color.b, color.a));
		disableTextures();
	}

	/** Returns the width of the given string for the given font. */
	public int getStringWidth(String msg, Font font) {
		ensureFontLoading(font);
		return fontMap.get(font).getWidth(msg);
	}
}
