package org.fortiss.tooling.kernel.ui.internal;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils.displayTimeConsumptionWarning;
import static org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils.triggerMarkerRefresh;
import static org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils.triggerMarkersRefresh;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.warning;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.resource.ImageDescriptor;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsPackage;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.IConstraintService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IConstraintUI;
import org.fortiss.tooling.kernel.ui.internal.introspection.items.ConstraintUIServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * Implementation of {@link IConstraintUIService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 23E1B2DF1FAB55250E64977BAC574B5F
 */
public final class ConstraintUIService implements IConstraintUIService, IIntrospectiveKernelService {

	/** The singleton service instance. */
	private static ConstraintUIService INSTANCE = new ConstraintUIService();

	/** Stores the string to class handler map. */
	protected final Map<String, Class<? extends IConstraint>> string2cstrMap =
			new HashMap<String, Class<? extends IConstraint>>();

	/** Stores the non-UI class to UI class handler map. */
	protected final Map<Class<? extends IConstraint>, Class<? extends IConstraintUI>> cstr2cstrUIMap =
			new HashMap<Class<? extends IConstraint>, Class<? extends IConstraintUI>>();

	/** Stores the UI-class to instance handler map. */
	protected final Map<Class<? extends IConstraintUI>, IConstraintUI> instanceMap =
			new HashMap<Class<? extends IConstraintUI>, IConstraintUI>();

	/** Returns the service instance. */
	public static ConstraintUIService getInstance() {
		return INSTANCE;
	}

	/** Activates logging to observe when constraints get outdated. */
	private static final boolean DEBUG = false;

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
		List<ITopLevelElement> tops = IPersistencyService.getInstance().getTopLevelElements();
		for(ITopLevelElement top : tops) {
			TreeIterator<EObject> tree = top.getRootModelElement().eAllContents();
			while(tree.hasNext()) {
				EObject next = tree.next();
				if(next instanceof ConstraintInstance) {
					install((ConstraintInstance)next);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void registerConstraintUI(Class<? extends IConstraintUI> cstrUIClass,
			Class<? extends IConstraint> cstrClass) {
		if(string2cstrMap.get(cstrClass) != null) {
			String name = cstrClass.getName();
			warning(ToolingKernelActivator.getDefault(),
					"Encountered more than one registered constraint UI associated with " + name);
		}
		string2cstrMap.put(cstrClass.getName(), cstrClass);
		cstr2cstrUIMap.put(cstrClass, cstrUIClass);
		try {
			instanceMap.put(cstrUIClass, cstrUIClass.getConstructor().newInstance());
		} catch(InstantiationException | IllegalAccessException | IllegalArgumentException |
				InvocationTargetException | NoSuchMethodException | SecurityException e) {
			warning(ToolingKernelActivator.getDefault(),
					"Instanciation of " + cstrUIClass.getName() + " threw an exception.");
		}
	}

	/** Initializes the service by setting up the handler map. */
	public void initializeService() {
		// nothing to do here
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service deals with the UI aspects of constraints." +
				"\n\nIt provides the following:" +
				"\n- inform whether a constraint instance can be open or not," +
				"\n- open the status of a constraint instance," +
				"\n- \"install\" a constraint instance, i.e., sets listeners on the constrained element(s) in order" +
				"\n  to automatically detect when a constraint is outdated," +
				"\n- inform, for a given constraint instance, whether it should display an error or a warning in case of non-satisfaction," +
				"\n- provide a user-friendly short description for a constraint (useful in various automatic messages)," +
				"\n- provide an icon for a given constraint (useful in the context menu)," +
				"\n- give the status of a given constraint instance - taking into account whether the constraint is activated or not," +
				"\n- activate a given constraint," +
				"\n- deactivate a given constraint," +
				"\n- give the list of all available constraints," +
				"\n- give the list of constraints which should always be activated - by opposition to those which should be explicitly activated by the user," +
				"\n- give the list of constraints which are currently activated - by the user or automatically," +
				"\n- complete a given project with instances of a given constraint if necessary." +
				"\n\nThis service is \"fed\" by classes of type IConstraintUI." +
				"\n\nThe class org.fortiss.tooling.kernel.ui.extension.base.ConstraintVerifierUIBaseAutocheck should" +
				"\ngenerally provide a sufficient initial implementation. You are however strongly encouraged" +
				"\nto extend this class to provide a better feedback to the user (e.g., with detailed errors)." +
				"\n\nSee the developer wiki page \"Add a new constraint on a metamodel\"" +
				" for detailed documentation on developing a new sort of constraint.";
	}

	/** @return the first registered constraint UI corresponding to <code>ci</code>. */
	private IConstraintUI getConstraintUI(ConstraintInstance ci) {
		Class<? extends IConstraint> cstr = string2cstrMap.get(ci.getConstraintName());
		Class<? extends IConstraintUI> cstrUI = cstr2cstrUIMap.get(cstr);
		return instanceMap.get(cstrUI);
	}

	/** {@inheritDoc} */
	@Override
	public void openStatus(ConstraintInstance ci) {
		getConstraintUI(ci).openStatus(ci, getStatus(ci));
		triggerMarkersRefresh(ci);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canOpen(ConstraintInstance ci) {
		IConstraintUI verifier = getConstraintUI(ci);
		return verifier.canOpen(ci, getStatus(ci));
	}

	/** Map containing the adapters of each constrained element. */
	Map<IConstrained, Adapter> constrainedAdapterMap = new HashMap<IConstrained, Adapter>();

	/** Installed constraints, sorted by adapters. */
	Map<Adapter, List<ConstraintInstance>> installedConstraintInstances = new HashMap<>();

	/** {@inheritDoc} */
	@Override
	public void install(ConstraintInstance ci) {
		triggerMarkersRefresh(ci);
		for(IConstrained constrained : ci.getConstraineds()) {
			Adapter currentAdapter = constrainedAdapterMap.get(constrained);
			if(currentAdapter == null) {
				currentAdapter = new EContentAdapter() {
					@Override
					public void notifyChanged(Notification notification) {
						super.notifyChanged(notification);
						if(isTouch(notification) || constraintRemoved(notification, constrained)) {
							return;
						}
						List<ConstraintInstance> outdatedConstraintInstances = new ArrayList<>();
						for(ConstraintInstance ci : installedConstraintInstances.get(this)) {
							if(getStatus(ci) != null) {
								constraintOutdated(ci, notification);
								outdatedConstraintInstances.add(ci);
							}
						}
						triggerMarkersRefresh(outdatedConstraintInstances);
					}
				};
				constrained.eAdapters().add(currentAdapter);
				constrainedAdapterMap.put(constrained, currentAdapter);
				installedConstraintInstances.put(currentAdapter, new ArrayList<>());
			}
			installedConstraintInstances.get(currentAdapter).add(ci);
		}
	}

	/**
	 * Checks if <code>notification</code> denotes a change entailing that <code>ci</code> is
	 * outdated. If so updates the status to outdated or triggers an automatic check if the
	 * <code>autocheck</code> field of the constraint is <code>true</code>.
	 * 
	 * @return <code>true</code> iff <code>notification</code> indeed entailed that <code>ci</code>
	 *         is outdated.
	 */
	private boolean constraintOutdated(ConstraintInstance ci, Notification notification) {
		if(skipStatusChangeNotification(ci, notification)) {
			return false;
		}
		if(!IConstraintService.getInstance().isUpToDate(ci)) {
			if(DEBUG) {
				String msg = "Constraint instance " + ci.toString() + " is outdated.";
				LoggingUtils.info(ToolingKernelUIActivator.getDefault(), msg);
			}
			if(!IConstraintService.getInstance().isUpdating(ci)) {
				IConstraintService.getInstance().markAsUpdating(ci);
				getConstraintUI(ci).onOutdate(ci);
				return true;
			}
		}
		return false;
	}

	/**
	 * @return true iff <code>notification</code> denotes a change of the <code>ci</code>'s
	 *         status.
	 */
	private boolean skipStatusChangeNotification(ConstraintInstance ci, Notification notification) {
		String n = ConstraintsPackage.eINSTANCE.getConstraintInstance_Status().getName();
		if(notification.getNotifier() != null && notification.getNotifier().equals(ci)) {
			if(notification.getFeature() instanceof EReference) {
				return(n.equals(((ENamedElement)notification.getFeature()).getName()));
			}
		}
		return false;
	}

	/**
	 * @return true if the notification is a consequence of a constraint of <code>constrained</code>
	 *         being removed.
	 */
	private boolean constraintRemoved(Notification notification, IConstrained constrained) {
		if(notification.getEventType() != Notification.REMOVE) {
			return false;
		}
		if(!notification.getNotifier().equals(constrained)) {
			return false;
		}
		if(notification.getFeature() instanceof EReference) {
			String name = ((ENamedElement)notification.getFeature()).getName();
			if(name.equals(ConstraintsPackage.eINSTANCE.getIConstrained_ConstraintInstances()
					.getName())) {
				Adapter cstrdAdapter = constrainedAdapterMap.get(constrained);
				List<ConstraintInstance> cis = installedConstraintInstances.get(cstrdAdapter);
				// To check!!! notification.getOldValue() should be the removed ConstraintInstance
				cis.remove(notification.getOldValue());
				if(cis.isEmpty()) {
					constrained.eAdapters().remove(cstrdAdapter);
					constrainedAdapterMap.remove(constrained);
				}
				return true;
			}
		}
		return true;
	}

	/** @return true if <code>notification</code> is not relevant and should be skipped */
	private static boolean isTouch(Notification notification) {
		// Notifications which do not notify about a model change.
		if(notification.isTouch()) {
			return true;
		}
		// Notifications initiated by the the AF3 kernel which are only GUI related.
		if(notification.getEventType() == EcoreUtils.KERNEL_EMF_EVENT_TYPE_COUNT) {
			return true;
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		Map<String, Class<? extends IConstraintUI>> handlerMap = new HashMap<>();
		for(Entry<String, Class<? extends IConstraint>> entry : string2cstrMap.entrySet()) {
			Class<? extends IConstraintUI> cstrUIClass = cstr2cstrUIMap.get(entry.getValue());
			handlerMap.put(entry.getKey(), cstrUIClass);
		}
		return new ConstraintUIServiceIntrospectionDetailsItem(handlerMap);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint UI Service";
	}

	/** Activated constraints per top level element. */
	private static Map<ITopLevelElement, Set<Class<? extends IConstraint>>> activatedConstraints =
			new HashMap<>();

	/** {@inheritDoc} */
	@Override
	public IConstraintInstanceStatus getStatus(ConstraintInstance ci) {
		ITopLevelElement top = IPersistencyService.getInstance().getTopLevelElementFor(ci);
		boolean isActivated = false;
		Set<Class<? extends IConstraint>> set = activatedConstraints.get(top);
		if(set != null) {
			for(Class<? extends IConstraint> id : set) {
				if(id.getName().equals(ci.getConstraintName())) {
					isActivated = true;
				}
			}
		}
		return isActivated ? ci.getStatus() : null;
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends IConstraint>> getAllConstraints() {
		return cstr2cstrUIMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends IConstraint>> getActivatedConstraints(ITopLevelElement top) {
		Set<Class<? extends IConstraint>> res = activatedConstraints.get(top);
		return res != null ? res : new HashSet<>();
	}

	/** {@inheritDoc} */
	@Override
	public void activate(Class<? extends IConstraint> cstrClass,
			IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		if(!activatedConstraints.containsKey(top)) {
			activatedConstraints.put(top, new HashSet<>());
		}
		activatedConstraints.get(top).add(cstrClass);
		List<ConstraintInstance> addeds =
				IConstraintService.getInstance().addMissingConstraintsInstances(top, cstrClass);
		IConstraintUI cstrUI = instanceMap.get(cstr2cstrUIMap.get(cstrClass));
		if(!addeds.isEmpty() && cstrUI.isTimeConsuming()) {
			displayTimeConsumptionWarning(cstrClass);
		}
		// Install notifiers for newly added constraints
		for(ConstraintInstance added : addeds) {
			IConstraintUIService.getInstance().install(added);
		}
		EList<ConstraintInstance> cstrs = cstrContainer.getConstraintInstances();
		Predicate<ConstraintInstance> isID = c -> cstrClass.getName().equals(c.getConstraintName());
		List<ConstraintInstance> idCstrs = cstrs.stream().filter(isID).collect(Collectors.toList());
		update(idCstrs);
	}

	/** {@inheritDoc} */
	@Override
	public void addMissingConstraintInstances(IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		List<ConstraintInstance> addeds = new ArrayList<ConstraintInstance>();
		for(Class<? extends IConstraint> cstrClass : getActivatedConstraints(top)) {
			addeds.addAll(IConstraintService.getInstance().addMissingConstraintsInstances(top,
					cstrClass));
			IConstraintService.getInstance().addMissingConstraintsInstances(top, cstrClass);
		}
		// Install notifiers for newly added constraints
		for(ConstraintInstance added : addeds) {
			IConstraintUIService.getInstance().install(added);
		}
		update(addeds);
	}

	/** Checks the constraints among the given list of constraint instances. */
	private void update(List<ConstraintInstance> cis) {
		// Checks all the constraints: the newly added and the already existing ones which are now
		// activated.
		for(ConstraintInstance ci : cis) {
			// No need to check if it turns out the status is already known from a previous
			// activation.
			if(!IConstraintService.getInstance().isUpToDate(ci) ||
					ci.getStatus() instanceof OutdatedConstraintInstanceStatus) {
				getConstraintUI(ci).onOutdate(ci);
			}
		}
		ConstraintsUIUtils.triggerMarkersRefresh(cis);
	}

	/** {@inheritDoc} */
	@Override
	public void deactivate(Class<? extends IConstraint> cstrClass,
			IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		activatedConstraints.get(top).remove(cstrClass);
		triggerMarkerRefresh(cstrClass, cstrContainer);
	}

	/** {@inheritDoc} */
	@Override
	public String getDescription(Class<? extends IConstraint> cstrClass) {
		IConstraintUI cstrUI = instanceMap.get(cstr2cstrUIMap.get(cstrClass));
		return cstrUI == null ? null : cstrUI.getDescription();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public ImageDescriptor getIconImageDescriptor(ConstraintInstance ci) {
		Class<? extends IConstraint> cstrClass = string2cstrMap.get(ci.getConstraintName());
		IConstraintUI cstrUI = instanceMap.get(cstr2cstrUIMap.get(cstrClass));
		return cstrUI == null ? null : cstrUI.getIconImageDescriptor();
	}

	/** {@inheritDoc} */
	@Override
	public boolean shallDisplayAsWarning(ConstraintInstance constraint) {
		return getConstraintUI(constraint).displayAsWarning();
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends IConstraint>> getAlwaysActivatedConstraints() {
		HashSet<Class<? extends IConstraint>> res = new HashSet<>();
		for(Entry<Class<? extends IConstraint>, Class<? extends IConstraintUI>> entry : cstr2cstrUIMap
				.entrySet()) {
			if(!instanceMap.get(entry.getValue()).shouldBeManuallyActivated()) {
				res.add(entry.getKey());
			}
		}
		return res;
	}

	/** {@inheritDoc} */
	@Override
	public String getDescription(ConstraintInstance ci) {
		Class<? extends IConstraint> cstrClass = string2cstrMap.get(ci.getConstraintName());
		return instanceMap.get(cstr2cstrUIMap.get(cstrClass)).getDescription(ci);
	}

	/** {@inheritDoc} */
	@Override
	public List<IFix> fixes(ConstraintInstance ci) {
		List<IFix> fixes = null;
		IConstraintUI cstr = getConstraintUI(ci);
		if(cstr != null) {
			fixes = cstr.fixes(ci, ci.getStatus());
		}
		return fixes != null ? fixes : new ArrayList<IFix>();
	}
}
