/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension.base;

import static org.fortiss.tooling.kernel.utils.ConstraintsUtils.createConstraintInstance;
import static org.fortiss.tooling.kernel.utils.ConstraintsUtils.createOutdatedStatus;
import static org.fortiss.tooling.kernel.utils.ConstraintsUtils.getConstrained;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.utils.ConstraintsUtils;

/**
 * Base for {@link IConstraint}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 23B91A136AE7587AA740DEF29F20EB13
 */
public class ConstraintBases {

	/** Most generic base. */
	public abstract static class ConstraintBase implements IConstraint {

		/** Returns the constraint instance container in the context of the given object. */
		abstract public IConstraintInstanceContainer getConstraintInstanceContainer(EObject obj);

		/** {@inheritDoc} */
		@Override
		public void preprocessBeforeChecksum(IConstrained cstrd) {
			cstrd.getConstraintInstances().clear();
			cstrd.getChecksumsPerConstraintName().clear();
			// By default, we remove all the constraint-related information.
			TreeIterator<EObject> it = cstrd.eAllContents();
			List<EMap<String, BigInteger>> toClears = new ArrayList<>();
			while(it.hasNext()) {
				EObject obj = it.next();
				if(obj instanceof IConstrained) {
					((IConstrained)obj).getConstraintInstances().clear();
					toClears.add(((IConstrained)obj).getChecksumsPerConstraintName());
				}
			}
			toClears.stream().forEach(toClear -> toClear.clear());
		}

		/** {@inheritDoc} */
		@Override
		public void cancel(ConstraintInstance ci) {
			// By default - and generally - nothing to do. Not all constraints are so heavy that
			// they deserve to have a cancellation procedure.
		}

		/** {@inheritDoc} */
		@Override
		public ConstraintInstance addConstraintInstanceIfNeeded(IConstrained constrained) {
			if(ConstraintsUtils.getConstraintInstanceOfType(constrained, this) != null) {
				return null;
			}
			ConstraintInstance ci = createConstraintInstanceIfNeeded(constrained);
			if(ci != null) {
				createOutdatedStatus(ci);
			}
			if(ci != null) {
				IConstraintInstanceContainer ciContainer =
						getConstraintInstanceContainer(constrained);
				ciContainer.getConstraintInstances().add(ci);
			}
			return ci;
		}

		/**
		 * Creates and return a constraint instance on the provided element.
		 * 
		 * Subclasses shall return <code>null</code> if the provided element should not be
		 * constrained (whatever is the reason: the constraint is only judge here).
		 */
		public abstract ConstraintInstance
				createConstraintInstanceIfNeeded(IConstrained constrained);
	}

	/**
	 * Base for a constraint which is as close as possible to the old constraint system:
	 * - the constrained element contains all the information necessary for the constraint
	 * (nothing stored in the constraint instance)
	 * - only one element is constrained
	 */
	public static abstract class ConstraintCheckerBase extends ConstraintBase {
		/** {@inheritDoc} */
		@Override
		public final IConstraintInstanceStatus verify(ConstraintInstance ci) {
			IConstrained constrained = getConstrained(ci);
			return isApplicable(constrained) ? verify(constrained) : null;
		}

		/** Verify the given constrained element. */
		public abstract IConstraintInstanceStatus verify(IConstrained constrained);

		/** Determines whether this constraint is applicable to the given model element. */
		public abstract boolean isApplicable(IConstrained constrained);

		/** {@inheritDoc} */
		@Override
		public ConstraintInstance addConstraintInstanceIfNeeded(IConstrained constrained) {
			return isApplicable(constrained) ? super.addConstraintInstanceIfNeeded(constrained)
					: null;
		}

		/** {@inheritDoc} */
		@Override
		public ConstraintInstance createConstraintInstanceIfNeeded(IConstrained constrained) {
			return createConstraintInstance(this.getClass(), constrained);
		}
	}
}
