/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.service;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConnectionCompositor;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.internal.ConnectionCompositorService;

/**
 * The connection compositor service provides registration and access to model
 * element connection compositors. A connection compositor knows how to create
 * and delete connections between two model elements. Here, a connection can be
 * a simple reference between two objects or, in a more complex case, an object
 * aggregated with some parent object and referencing two other objects in the
 * model element tree.
 * 
 * @see IConnectionCompositor
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: BF0A98CF3E3B3D918F34709867076AC6
 */
public interface IConnectionCompositorService {
	/** Returns the service instance. */
	public static IConnectionCompositorService getInstance() {
		return ConnectionCompositorService.getInstance();
	}

	/**
	 * Returns whether a connection between the source and the target is
	 * possible for the given parent. Additional information can be included in
	 * the context, which also may be null.
	 */
	boolean canConnect(EObject source, EObject target, EObject parent,
			IConnectionCompositionContext context);

	/**
	 * Actually connects the given source and target. This will only be called
	 * if {@link #canConnect(EObject, EObject, EObject, IConnectionCompositionContext)} returned
	 * true.
	 * 
	 * @param source
	 *            the source of the connection.
	 * @param target
	 *            the target of the connection.
	 * @param parent
	 *            the parent of the connection.
	 * @param context
	 *            context information for the connection. May be null.
	 * 
	 * @return {@code true} iff the connect operation has successfully been performed.
	 */
	boolean connect(EObject source, EObject target, EObject parent,
			IConnectionCompositionContext context);

	/**
	 * Determines if there is a compositor that allows the reconnection of the
	 * given connection to the given source and target.
	 */
	boolean canReconnect(EObject connection, EObject source, EObject target,
			IConnectionCompositionContext context);

	/**
	 * Reconnects the connection. Since this operation may be canceled by the
	 * user, it returns a boolean value.
	 */
	boolean reconnect(EObject connection, EObject source, EObject target,
			IConnectionCompositionContext context);

	/**
	 * Determines if there is a compositor that allows the decomposition of the
	 * given connection.
	 */
	boolean canDisconnect(EObject connection);

	/**
	 * Decomposes the connection. Since this operation may be canceled by the
	 * user, it returns a boolean value.
	 */
	boolean disconnect(EObject connection);

	/** Registers the given connection compositor with the service. */
	void registerConnectionCompositor(IConnectionCompositor<EObject, EObject, EObject> compositor,
			Class<?> sourceClass, Class<?> targetClass);
}
