/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.treemap;

import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackAdapter;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;

import org.conqat.lib.commons.assertion.CCSMPre;
import org.conqat.lib.commons.treemap.FlatTreeMapRenderer;
import org.conqat.lib.commons.treemap.ITreeMapLayoutAlgorithm;
import org.conqat.lib.commons.treemap.ITreeMapNode;
import org.conqat.lib.commons.treemap.ITreeMapRenderer;
import org.conqat.lib.commons.treemap.NodeTextRenderer;
import org.conqat.lib.commons.treemap.StripeTreeMapAlgorithm;
import org.conqat.ide.commons.ui.color.ConstantColorProvider;
import org.conqat.ide.commons.ui.color.IColorProvider;
import org.conqat.ide.commons.ui.swt.SWTUtils;

/**
 * The SWT widget representing the TreeMap
 * 
 * @author Timo Besenreuther
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: 0EA89CFD61F2EFEF0508738220FEFA23
 */

public class TreeMapWidget<T> extends Canvas {

	/** Default color used */
	private static final Color DEFAULT_COLOR = new Color(null, new RGB(128,
			128, 128));

	/** The input */
	private T input;

	/** The root of the node tree */
	private ProviderTreeMapNode<T> rootNode;

	/** Content provider. */
	private ITreeMapContentProvider<T> contentProvider;

	/** Color provider. */
	private IColorProvider<T> colorProvider = new ConstantColorProvider<T>(
			DEFAULT_COLOR);

	/** Pattern provider. */
	private IPatternProvider<T> patternProvider = null;

	/** Label provider. */
	private ITreeMapLabelProvider<T> labelProvider = new TreeMapLabelProviderBase<T>();

	/** Treemap renderer. */
	private ITreeMapRenderer renderer = new FlatTreeMapRenderer();

	/** Treemap layout algorithm. */
	private ITreeMapLayoutAlgorithm algorithm = new StripeTreeMapAlgorithm();

	/** Text color. If null, no text will be shown */
	private java.awt.Color textColor = null;

	/** The node the mouse is currently on */
	private ProviderTreeMapNode<T> mouseOnNode = null;

	/** Cache the treemap image */
	private Image treeMapCache = null;

	/** List of listeners */
	private final List<ITreeMapListener<T>> listeners = new ArrayList<ITreeMapListener<T>>();

	/**
	 * Pattern used to split tree map node names that are too long to be
	 * displayed into parts from which the last part is chosen for display (if
	 * this is null, name is not split)
	 */
	private Pattern separationPattern = null;

	/** Constructor. */
	public TreeMapWidget(Composite parent) {
		super(parent, SWT.NONE);

		// resize listener
		addControlListener(new ControlAdapter() {
			/** {@inheritDoc} */
			@Override
			public void controlResized(ControlEvent e) {
				refresh();
			}
		});

		// paint listener
		addPaintListener(new PaintListener() {
			/** {@inheritDoc} */
			public void paintControl(PaintEvent e) {
				paint(e);
			}
		});

		// mouse listeners
		addMouseMoveListener(new MouseMoveListener() {
			/** {@inheritDoc} */
			@Override
			public void mouseMove(MouseEvent e) {
				mouseMoved(e);
			}
		});
		addMouseListener(new MouseAdapter() {
			/** {@inheritDoc} */
			@Override
			public void mouseDoubleClick(MouseEvent e) {
				doubleClicked(e);
			}

			/** {@inheritDoc} */
			@Override
			public void mouseDown(MouseEvent e) {
				clicked(e);
			}
		});
		addMouseTrackListener(new MouseTrackAdapter() {
			@Override
			public void mouseExit(MouseEvent e) {
				mouseExitMap();
			}
		});
	}

	/** Sets the content provider. */
	public void setContentProvider(ITreeMapContentProvider<T> contentProvider) {
		CCSMPre.isNotNull(contentProvider);
		this.contentProvider = contentProvider;
		refresh();
	}

	/** Returns the content provider */
	/* package */ITreeMapContentProvider<T> getContentProvider() {
		return contentProvider;
	}

	/** Sets the color provider. */
	public void setColorProvider(IColorProvider<T> colorProvider) {
		CCSMPre.isNotNull(colorProvider);
		this.colorProvider = colorProvider;
		refresh();
	}

	/** Returns the color provider */
	/* package */IColorProvider<T> getColorProvider() {
		return colorProvider;
	}

	/** Sets the pattern provider. */
	public void setPatternProvider(IPatternProvider<T> patternProvider) {
		CCSMPre.isNotNull(patternProvider);
		this.patternProvider = patternProvider;
		refresh();
	}

	/** Returns the pattern provider */
	/* package */IPatternProvider<T> getPatternProvider() {
		return patternProvider;
	}

	/** Sets the label provider. */
	public void setLabelProvider(ITreeMapLabelProvider<T> labelProvider) {
		CCSMPre.isNotNull(labelProvider);
		this.labelProvider = labelProvider;
		refresh();
	}

	/** Returns the label provider */
	/* package */ITreeMapLabelProvider<T> getLabelProvider() {
		return labelProvider;
	}

	/** Sets the text color */
	public void setTextColor(Color color) {
		CCSMPre.isNotNull(color);
		this.textColor = new java.awt.Color(color.getRed(), color.getGreen(),
				color.getBlue());
		refresh();
	}

	/** Sets the layout algorithm */
	public void setLayoutAlgorithm(ITreeMapLayoutAlgorithm algorithm) {
		CCSMPre.isNotNull(algorithm);
		this.algorithm = algorithm;
		refresh();
	}

	/** Add a listener */
	public void addTreeMapListener(ITreeMapListener<T> listener) {
		listeners.add(listener);
	}

	/** Remove a listener */
	public void removeTreeMapLister(ITreeMapListener<T> listener) {
		listeners.remove(listener);
	}

	/** Sets the separation pattern. */
	public void setSeparationPattern(Pattern separationPattern) {
		CCSMPre.isNotNull(separationPattern);
		this.separationPattern = separationPattern;
		refresh();
	}

	/** Sets the renderer */
	public void setRenderer(ITreeMapRenderer renderer) {
		CCSMPre.isNotNull(renderer);
		this.renderer = renderer;
		refresh();
	}

	/** Returns the renderer */
	/* package */ITreeMapRenderer getRenderer() {
		return renderer;
	}

	/** Sets the input forces a redraw. */
	public void setInput(T input) {
		this.input = input;
		refresh();
	}

	/** Delete all caches and redraw */
	public void refresh() {
		mouseOnNode = null;
		if (treeMapCache != null) {
			treeMapCache.dispose();
		}
		treeMapCache = null;
		redraw();
	}

	/** Paint the bar */
	private void paint(PaintEvent event) {
		GC context = event.gc;
		Display display = getDisplay();

		Rectangle rect = getClientArea();
		context.fillRectangle(rect);
		drawBorder(context, rect.x, rect.y, rect.width - 1, rect.height - 1,
				display.getSystemColor(SWT.COLOR_WIDGET_NORMAL_SHADOW), display
						.getSystemColor(SWT.COLOR_WIDGET_HIGHLIGHT_SHADOW));

		drawTreeMap(context, rect.width - 3, rect.height - 3);

		if (mouseOnNode != null) {
			highlightNode(context, mouseOnNode, true);
		}
	}

	/** Draw the border of the progress bar */
	private void drawBorder(GC gc, int x, int y, int w, int h, Color topleft,
			Color bottomright) {
		gc.setForeground(topleft);
		gc.drawLine(x, y, x + w - 1, y);
		gc.drawLine(x, y, x, y + h - 1);

		gc.setForeground(bottomright);
		gc.drawLine(x + w, y, x + w, y + h);
		gc.drawLine(x, y + h, x + w, y + h);
	}

	/** Draw the TreeMap image. */
	private synchronized void drawTreeMap(GC context, int width, int height) {
		if (treeMapCache == null) {
			mouseOnNode = null;
			BufferedImage image = new BufferedImage(width, height,
					BufferedImage.TYPE_INT_RGB);
			Graphics2D graphics = image.createGraphics();

			rootNode = new ProviderTreeMapNode<T>(input, this, null);

			// layout map
			algorithm.layout(rootNode, new java.awt.Rectangle(0, 0, width,
					height));

			// render map
			ITreeMapRenderer renderer = getRenderer();
			renderer.renderTreeMap(rootNode, graphics);

			// render text
			if (textColor != null) {
				NodeTextRenderer texter = new NodeTextRenderer(textColor,
						separationPattern);
				texter.renderTreeMap(rootNode, graphics);
			}

			ImageData imageData = SWTUtils.convertToSWT(image);
			treeMapCache = new Image(getDisplay(), imageData);
		}
		context.drawImage(treeMapCache, 0, 0, width, height, 1, 1, width,
				height);
	}

	/** Highlight a node */
	private void highlightNode(GC context, ProviderTreeMapNode<T> node,
			boolean mainHighlight) {
		Rectangle2D nodeRect = node.getLayoutRectangle();
		Rectangle rect = new Rectangle((int) nodeRect.getX() + 1,
				(int) nodeRect.getY() + 1, (int) nodeRect.getWidth(),
				(int) nodeRect.getHeight());
		context.setLineWidth(2);
		context.drawRectangle(rect);

		if (mainHighlight) {
			context.setAlpha(100);
			context.fillRectangle(rect);
			context.setAlpha(255);
		}

		ProviderTreeMapNode<T> parent = node.getParent();
		if (parent != null && parent.getParent() != null) {
			highlightNode(context, parent, false);
		}
	}

	/** Callback for mouse move event */
	private synchronized void mouseMoved(MouseEvent e) {
		ProviderTreeMapNode<T> node = (ProviderTreeMapNode<T>) findNode(
				rootNode, new Point(e.x, e.y));
		if (node != mouseOnNode) {
			mouseOnNode = node;
			redraw();
			for (ITreeMapListener<T> listener : listeners) {
				listener.mouseEnterElement(mouseOnNode.getUserDatum());
			}
		}
	}

	/** Callback for mouse exit event */
	private synchronized void mouseExitMap() {
		// remove highlight
		mouseOnNode = null;
		redraw();
		for (ITreeMapListener<T> listener : listeners) {
			listener.mouseExitElement();
		}
	}

	/** Callback for click event */
	private synchronized void clicked(MouseEvent e) {
		ProviderTreeMapNode<T> node = (ProviderTreeMapNode<T>) findNode(
				rootNode, new Point(e.x, e.y));
		for (ITreeMapListener<T> listener : listeners) {
			listener.clickedElement(node.getUserDatum(), e);
		}
	}

	/** Callback for double click event */
	private synchronized void doubleClicked(MouseEvent e) {
		ProviderTreeMapNode<T> node = (ProviderTreeMapNode<T>) findNode(
				rootNode, new Point(e.x, e.y));
		for (ITreeMapListener<T> listener : listeners) {
			listener.doubleClickedElement(node.getUserDatum(), e);
		}
	}

	/** Find a node */
	private ITreeMapNode<T> findNode(ITreeMapNode<T> parent, Point point) {
		List<ITreeMapNode<T>> children = parent.getChildren();
		if (children == null || children.size() == 0) {
			return parent;
		}
		for (ITreeMapNode<T> child : children) {
			if (child.getLayoutRectangle().contains(point)) {
				return findNode(child, point);
			}
		}
		return null;
	}

	/** Dispose widget. */
	@Override
	public void dispose() {
		if (treeMapCache != null) {
			treeMapCache.dispose();
		}
		super.dispose();
	}

	/**
	 * Save the displayed image to a file using a certain format. See
	 * {@link ImageLoader#save} for the supported formats.
	 */
	public void saveToFile(String filename, int format) {
		CCSMPre.isNotNull(treeMapCache);
		ImageLoader imageLoader = new ImageLoader();
		imageLoader.data = new ImageData[] { treeMapCache.getImageData() };
		imageLoader.save(filename, format);
	}
}